"""Top-level package for meteodata-lab."""

# Standard library
import importlib.metadata

__author__ = "Carlos Osuna"
__email__ = "carlos.osuna@meteoswiss.ch"
__version__ = importlib.metadata.version("meteodata-lab")
